//////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2007-2017 musikcube team
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//    * Redistributions of source code must retain the above copyright notice,
//      this list of conditions and the following disclaimer.
//
//    * Redistributions in binary form must reproduce the above copyright
//      notice, this list of conditions and the following disclaimer in the
//      documentation and/or other materials provided with the distribution.
//
//    * Neither the name of the author nor the names of other contributors may
//      be used to endorse or promote products derived from this software
//      without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
//////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"

#include <stdlib.h>
#include <time.h>

#include <cstdio>

#include <cursespp/App.h>
#include <cursespp/Screen.h>

#include <app/layout/MainLayout.h>
#include <app/util/GlobalHotkeys.h>
#include <app/util/Hotkeys.h>
#include <app/util/PreferenceKeys.h>

#include <core/i18n/Locale.h>
#include <core/audio/PlaybackService.h>

#include <glue/audio/MasterTransport.h>

#include <core/debug.h>
#include <core/library/LibraryFactory.h>
#include <core/plugin/Plugins.h>
#include <core/support/PreferenceKeys.h>
#include <core/audio/Visualizer.h>
#include <core/support/Common.h>

#include <boost/locale.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/fstream.hpp>
#include <boost/filesystem/detail/utf8_codecvt_facet.hpp>

#ifdef WIN32
    #include <cursespp/Win32Util.h>
    #include "resource.h"
    #undef MOUSE_MOVED
#endif

using namespace musik;
using namespace musik::glue;
using namespace musik::glue::audio;
using namespace musik::core;
using namespace musik::core::audio;
using namespace musik::cube;
using namespace cursespp;

#define MIN_WIDTH 54
#define MIN_HEIGHT 12

#ifdef WIN32
int _main(int argc, wchar_t* argv[]);

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int nCmdShow) {
    return App::Running("musikcube") ? 0 : _main(0, 0);
}

int _main(int argc, wchar_t* argv[]) {
#else
int main(int argc, char* argv[]) {
#endif

    srand((unsigned int) time(0));

    /* the following allows boost::filesystem to use utf8 on Windows */
    std::locale locale = std::locale();
    std::locale utf8Locale(locale, new boost::filesystem::detail::utf8_codecvt_facet);
    boost::filesystem::path::imbue(utf8Locale);

    /* ensure we have the correct locale loaded */
    musik::core::i18n::Locale::Instance().Initialize(musik::core::GetApplicationDirectory() + "/locales/");

#ifdef WIN32
    musik::core::RemoveOldDlls();
    AddDllDirectory(u8to16(musik::core::GetPluginDirectory()).c_str());
#endif

#ifndef WIN32
    #if 1 /*DEBUG*/
        freopen("/tmp/musikcube.log", "w", stderr);
    #else
        freopen("/dev/null", "w", stderr);
    #endif
#endif

#ifdef __PDCURSES__
    PDC_set_resize_limits(MIN_HEIGHT, 1000, MIN_WIDTH, 1000);
    resize_term(26, 100); /* must be before app init */
#endif

    musik::core::MigrateOldDataDirectory();

    musik::debug::init();

    ILibraryPtr library = LibraryFactory::Libraries().at(0);
    library->SetMessageQueue(Window::MessageQueue());

    auto prefs = Preferences::ForComponent(
        musik::core::prefs::components::Settings);

    MasterTransport transport;
    PlaybackService playback(Window::MessageQueue(), library, transport);

    GlobalHotkeys globalHotkeys(playback, library);
    Window::SetNavigationKeys(Hotkeys::NavigationKeys());

    musik::core::plugin::InstallDependencies(library);

    {
        App app("musikcube"); /* must be before layout creation */

#ifdef WIN32
        app.SetIcon(IDI_ICON1);
        app.SetSingleInstanceId("musikcube");
#endif
        /* fire up the indexer if configured to run on startup */
        if (prefs->GetBool(musik::core::prefs::keys::SyncOnStartup, true)) {
            library->Indexer()->Schedule(IIndexer::SyncType::All);
        }

        /* set color mode (basic, palette, rgb) */
        Colors::Mode colorMode = Colors::RGB;

        if (prefs->GetBool(musik::cube::prefs::keys::DisableCustomColors.c_str(), false)) {
            colorMode = Colors::Basic;
        }
#ifndef WIN32
        else if (prefs->GetBool(musik::cube::prefs::keys::UsePaletteColors.c_str(), true)) {
            colorMode = Colors::Palette;
        }
#endif

        app.SetColorMode(colorMode);

        /* theme */
        std::string colorTheme = prefs->GetString(musik::cube::prefs::keys::ColorTheme);
        if (colorTheme.size()) {
            app.SetColorTheme(colorTheme);
        }

        /* tray icon */
        app.SetMinimizeToTray(prefs->GetBool(musik::cube::prefs::keys::MinimizeToTray, false));
        if (prefs->GetBool(musik::cube::prefs::keys::StartMinimized, false)) {
            app.Minimize();
        }

        app.SetMinimumSize(MIN_WIDTH, MIN_HEIGHT);

        /* main layout */
        using Main = std::shared_ptr<MainLayout>;
        Main mainLayout(new MainLayout(app, playback, transport, library));

        mainLayout->Start();

        app.SetKeyHandler([&](const std::string& kn) -> bool {
            if (app.IsOverlayVisible()) {
                return false;
            }

            return globalHotkeys.Handle(kn);
        });

        /* blocking event loop */
        app.Run(mainLayout);
        /* done with the app */

        mainLayout->Stop();

#ifdef WIN32
        win32::HideMainWindow();
#endif
    }

    musik::core::audio::vis::HideSelectedVisualizer();
    musik::core::plugin::UninstallDependencies();

    LibraryFactory::Instance().Shutdown();

    musik::debug::deinit();

    return 0;
}
